from spinn_front_end_common.abstract_models\
    .abstract_send_me_multicast_commands_vertex \
    import AbstractSendMeMulticastCommandsVertex
from spinn_front_end_common.utilities.database.database_connection \
    import DatabaseConnection


class EthernetCommandConnection(DatabaseConnection):
    """ A connection that can send commands to a device at the start and end\
        of a simulation
    """

    def __init__(
            self, translator, command_containers=None, local_host=None,
            local_port=19999):
        """

        :param translator:\
            A translator of multicast commands to device commands
        :param command_containers:\
            A list of instances of AbstractSendMeMulticastCommandsVertex that\
            have commands to be sent at the start and end of simulation
        :param local_host:\
            The optional host to listen on for the start/resume message
        :param local_port:\
            The optional port to listen on for the stop/pause message
        """

        DatabaseConnection.__init__(
            self, start_resume_callback_function=self._start_resume_callback,
            stop_pause_callback_function=self._stop_pause_callback,
            local_host=local_host, local_port=local_port)

        self._command_containers = list()
        if command_containers is not None:
            for command_container in command_containers:
                self.add_command_container(command_container)
        self._translator = translator

    def add_command_container(self, command_container):
        """ Add a command container

        :param command_container:\
            An instance of AbstractSendMeMulticastCommandsVertex that\
            has commands to be sent at the start and end of simulation
        """
        if not isinstance(
                command_container, AbstractSendMeMulticastCommandsVertex):
            raise Exception(
                "Each command container must be an instance of"
                " AbstractSendMeMulticastCommandsVertex")
        if len(command_container.timed_commands) > 0:
            raise Exception("Timed commands cannot be handled by this class")
        self._command_containers.append(command_container)

    def _start_resume_callback(self):

        # Send commands from each command container
        for command_container in self._command_containers:
            for command in command_container.start_resume_commands:
                self._translator.translate_control_packet(command)

    def _stop_pause_callback(self):

        # Send commands from each command container
        for command_container in self._command_containers:
            for command in command_container.pause_stop_commands:
                self._translator.translate_control_packet(command)
